#! /usr/bin/env python

### Midnight Philander, a minimal file commander
### Giuseppe Cocomazzi 2016 sbudella@gmail.com

import os
import sys
import stat
import shutil
import curses
import curses.textpad

from itertools import islice
from datetime import datetime
from time import strftime
from pwd import getpwuid
from grp import getgrgid


fmt_fns = [
    lambda st: {
        stat.S_IFREG: 'f',
        stat.S_IFBLK: 'b',
        stat.S_IFCHR: 'c',
        stat.S_IFDIR: 'd',
        stat.S_IFLNK: 'l',
        stat.S_IFIFO: 'p',
        stat.S_IFSOCK: 's',
    }.get(st.st_mode & stat.S_IFMT(st.st_mode), 'f'),
    lambda st: oct(st.st_mode & stat.S_IRWXU),
    lambda st: str(st.st_nlink),
    lambda st: getpwuid(st.st_uid).pw_name + ' ' + getgrgid(st.st_gid).gr_name,
    lambda st: str(st.st_size),
    lambda st: strftime('%b %d %Y %H:%M',
        datetime.fromtimestamp(st.st_mtime).timetuple()),
]

def file_info(path):
    st = os.lstat(path)
    base = ' '.join(map(lambda f: f(st), fmt_fns)) + ' ' + os.path.basename(path)
    info = base if base[0] != 'l' else base + ' -> ' + os.readlink(path)
    return info if len(info) < w else info[:w - 3] + '..'


class Directory(object):
    def __init__(self, d, h, start=0, stop=0, last=0):
        self.root = os.path.abspath(d)
        self.name = d
        self.dirs = ['..'] + sorted(os.listdir(d))
        self.len = len(self.dirs)
        self.h = h if h < self.len else self.len
        self.start = start if start > 0 else 0
        self.stop = stop if stop > 0 else h
        self.last = self.y = last if last >= 0 else self.h - 1

    def __getitem__(self, key):
        return self.dirs[key]


def init():
    win = curses.initscr()
    curses.noecho()
    curses.cbreak()
    win.keypad(True)
    win.clear()

    return win


def exit(win):
    curses.nocbreak()
    win.keypad(0)
    curses.echo()
    curses.endwin()


def draw(ldir, rdir, y=0, x=0, curdir=None):
    if not curdir:
        curdir = ldir
    pane = int(x > w/2)
    win.clear()

    for d, p in ((ldir, 0), (rdir, w/2 + 1)):
        for i, name in enumerate(islice(d, d.start, d.stop)):
            if os.path.isdir(os.path.join(d.root, name)):
                attr = curses.A_BOLD
            else:
                attr = 0
            win.addstr(i, p,
                       name if len(name) < w/2 else name[:w/2 - 3] + '..',
                       attr)

    win.vline(0, w/2, '|', h)
    last = curdir[curdir.last]
    win.addstr(y, pane * (w/2 + 1),
               last if len(last) < w/2 else last[:w/2 - 3] + '..',
               curses.A_REVERSE)
    win.refresh()
    write_status(' | '.join((
        "F4 Details",
        "F5 Copy",
        "F6 Move",
        "F7 Rename", 
        "F8 Delete", 
        "F9 Mkdir",
        "F10 Quit"
    )))


def textbox(y=None, x=None):
    if y is not None and x is not None: 
        pane = int(x > w/2)
        return curses.textpad.Textbox(
            curses.newwin(1, w/2, y, pane * (w/2 + 1))
        ).edit()
    else:
        status_win.clear()
        status_win.refresh()
        return curses.textpad.Textbox(status_win).edit()


def write_status(msg):
    status_win.clear()
    status_win.addstr(0, 0, msg, curses.A_REVERSE)
    status_win.refresh()


# boycott shutil.copytree
def copy(src, dst):
    tgt = os.path.join(dst, os.path.basename(src))
    if os.path.isdir(src):
        os.mkdir(tgt)
        for elm in os.listdir(src):
            copy(os.path.join(src, elm), tgt)
    else:
        shutil.copy2(src, tgt)


if __name__ == '__main__':
    win = init()
    h, w = win.getmaxyx()
    h -= 1

    ldir, rdir = dirs = [Directory(d, h)
                         for d in (sys.argv[1:] + [os.getcwd(),]*2)][0:2]
    hs = ldir.h, rdir.h

    status_win = curses.newwin(1, w, h, 0)

    draw(ldir, rdir, ldir.last, 0)
    y, x = win.getyx()
    pane = int(x > w/2)
    c = False

    while c != curses.KEY_F10:
        try:
            curdir = dirs[pane]
            selected = os.path.join(curdir.root, curdir[curdir.last])
            curdir.y = y
            backdir = dirs[not pane]
            curh = hs[pane] 
            c = win.getch()

            ### Scroll down
            if c == curses.KEY_DOWN:
                if curdir.last < curdir.len - 1:
                    curdir.last += 1
                    if y < curh - 1:
                        y += 1
                        draw(ldir, rdir, y, x, curdir)
                    else:
                        y = curh - 1
                        if curdir.start + 1 < curdir.len:
                            curdir.start += 1
                            curdir.stop += 1
                            draw(ldir, rdir, y, x, curdir)

            ### Scroll up
            elif c == curses.KEY_UP:
                if curdir.last > 0:
                    curdir.last -= 1
                    if y > 0:
                        y -= 1
                        draw(ldir, rdir, y, x, curdir)
                    else:
                        y = 0 
                        if curdir.start:
                            curdir.start -= 1
                            curdir.stop -= 1
                            draw(ldir, rdir, y, x, curdir)

            ### Switch pane
            elif c == ord('\t'):
                if pane:
                    x -= w/2 + 1
                else:
                    x += w/2 + 1
                pane = not pane
                curdir = dirs[pane]
                y = curdir.y
                draw(ldir, rdir, y, x, curdir)

            ### Enter directory
            elif c == 10:
                if os.path.isdir(selected):
                    curdir = dirs[pane] = Directory(selected, h)
                    ldir, rdir = dirs
                    hs = ldir.h, rdir.h
                    draw(ldir, rdir, curdir.last, x, curdir)
                    y, x = win.getyx()

            ### Copy selection
            elif c == curses.KEY_F5:
                if selected[-2:] == '..':
                    raise OSError('Cannot copy parent directory.')
                if backdir.root == curdir.root:
                    raise OSError('Destination already exists.')

                copy(selected, backdir.root)
                dirs[not pane] = Directory(backdir.root, h)
                ldir, rdir = dirs
                hs = ldir.h, rdir.h
                draw(ldir, rdir, y, x, curdir)
                y, x = win.getyx()

            ### Remove selection
            elif c == curses.KEY_F8:
                if selected[-2:] == '..':
                    raise OSError('Cannot remove parent directory.')

                if os.path.isdir(selected) and not os.path.islink(selected):
                    shutil.rmtree(selected)
                else:
                    os.unlink(selected)
                curdir = dirs[pane] = Directory(curdir.root, h, curdir.start,
                                                curdir.stop, curdir.last - 1)
                if backdir.root == curdir.root:
                    dirs[not pane] = curdir
                ldir, rdir = dirs
                hs = ldir.h, rdir.h
                draw(ldir, rdir, y - 1, x, curdir)
                y, x = win.getyx()

            ### Rename selection
            elif c == curses.KEY_F7:
                if selected[-2:] == '..':
                    raise OSError('Cannot rename parent directory.')

                newname = textbox(y, x).strip()
                if newname == '..':
                    raise OSError('Cannot rename to parent directory.')
                shutil.move(selected, os.path.join(curdir.root, newname))
                curdir = dirs[pane] = Directory(curdir.root, h, curdir.start,
                                                curdir.stop, curdir.last)
                if backdir.root == curdir.root:
                    dirs[not pane] = curdir
                ldir, rdir = dirs
                hs = ldir.h, rdir.h
                draw(ldir, rdir, y, x, curdir)

            ### Move selection
            elif c == curses.KEY_F6:
                if selected[-2:] == '..':
                    raise OSError('Cannot move parent directory.')

                shutil.move(selected, backdir.root)
                dirs[pane], dirs[not pane] = ( 
                    Directory(curdir.root, h, curdir.start, curdir.stop,
                              curdir.last - 1),
                    Directory(backdir.root, h, backdir.start, backdir.stop,
                              backdir.last))
                curdir = dirs[pane]
                ldir, rdir = dirs
                hs = ldir.h, rdir.h
                draw(ldir, rdir, y - 1, x, curdir)
                y, x = win.getyx()

            ### Make directory
            elif c == curses.KEY_F9:
                name = textbox().strip()
                if name == '..':
                    raise OSError('Cannot rename to parent directory.')
                os.mkdir(os.path.join(curdir.root, name))
                curdir = dirs[pane] = Directory(curdir.root, h, curdir.start,
                                                curdir.stop, curdir.last)
                ldir, rdir = dirs
                hs = ldir.h, rdir.h
                draw(ldir, rdir, y, x, curdir)

            ### File info
            elif c == curses.KEY_F4:
                write_status(file_info(selected))

        except Exception, e:
            write_status(e.message)

    exit(win)
